import pandas as pd
import pytest

from pvoutput import mapscraper as ms
from pvoutput.consts import MAP_URL


def compare_function_output_to_pickle(key, function_dict, dict_of_dfs, series=False):
    df_from_func = function_dict[key]()
    test_df = dict_of_dfs[key]
    if series:
        return pd.testing.assert_series_equal(df_from_func, test_df)
    return pd.testing.assert_frame_equal(df_from_func, test_df, check_like=True)


def test_convert_to_country_code():
    assert ms._convert_to_country_code(1) == 1
    assert ms._convert_to_country_code("United Kingdom") == 243

    def _assert_raises(bad_countries, exception):
        for bad_country in bad_countries:
            with pytest.raises(exception):
                ms._convert_to_country_code(bad_country)
                pytest.fail(
                    "Failed to raise {} for country={}".format(exception.__name__, bad_country)
                )

    _assert_raises([-1, -100, 1000, "blah"], ValueError)


def test_create_map_url():
    assert ms._create_map_url() == MAP_URL
    assert ms._create_map_url(country_code=1) == MAP_URL + "?country=1"
    assert ms._create_map_url(page_number=2) == MAP_URL + "?p=2"
    assert ms._create_map_url(ascending=True) == MAP_URL + "?d=asc"
    assert ms._create_map_url(ascending=False) == MAP_URL + "?d=desc"
    assert ms._create_map_url(sort_by="efficiency") == MAP_URL + "?o=gss"
    with pytest.raises(ValueError):
        ms._create_map_url(sort_by="blah")


def test_pv_system_size_metadata(get_function_dict, get_test_dict_of_dfs):
    assert (
        compare_function_output_to_pickle(
            "pv_system_size_metadata", get_function_dict, get_test_dict_of_dfs
        )
        is None
    )


def test_process_output_col(get_function_dict, get_test_dict_of_dfs):
    assert (
        compare_function_output_to_pickle(
            "process_output_col", get_function_dict, get_test_dict_of_dfs, series=True
        )
        is None
    )


def test_process_generation_and_average_cols(get_function_dict, get_test_dict_of_dfs):
    assert (
        compare_function_output_to_pickle(
            "process_generation_and_average_cols", get_function_dict, get_test_dict_of_dfs
        )
        is None
    )


def test_process_efficiency_col(get_function_dict, get_test_dict_of_dfs):
    assert (
        compare_function_output_to_pickle(
            "process_efficiency_col", get_function_dict, get_test_dict_of_dfs, series=True
        )
        is None
    )


def test_process_metadata(get_function_dict, get_test_dict_of_dfs):
    assert (
        compare_function_output_to_pickle(
            "process_metadata", get_function_dict, get_test_dict_of_dfs
        )
        is None
    )



import inspect
import os
import pickle
from functools import partial

import pytest

from pvoutput import mapscraper as ms


@pytest.fixture
def data_dir():
    # Taken from http://stackoverflow.com/a/6098238/732596
    data_dir = os.path.dirname(inspect.getfile(inspect.currentframe()))
    data_dir = os.path.abspath(data_dir)
    assert os.path.isdir(data_dir), data_dir + " does not exist."
    return data_dir


def get_cleaned_test_soup(data_dir):
    test_soup_filepath = os.path.join(data_dir, "data/mapscraper_soup.pickle")
    with open(test_soup_filepath, "rb") as f:
        test_soup = pickle.load(f)
    return ms.clean_soup(test_soup)


@pytest.fixture()
def get_test_dict_of_dfs(data_dir):
    dict_filepath = os.path.join(data_dir, "data/mapscraper_dict_of_dfs.pickle")
    with open(dict_filepath, "rb") as f:
        test_soup = pickle.load(f)
    return test_soup


@pytest.fixture()
def get_function_dict(data_dir):
    # using partials so functions only get executed when needed
    soup = get_cleaned_test_soup(data_dir)
    df = ms._process_system_size_col(soup)
    index = df.index
    keys = get_keys_for_dict()
    functions = (
        partial(ms._process_system_size_col, soup),
        partial(ms._process_output_col, soup, index),
        partial(ms._process_generation_and_average_cols, soup, index),
        partial(ms._process_efficiency_col, soup, index),
        partial(ms._process_metadata, soup),
    )
    function_dict = dict(zip(keys, functions))
    return function_dict


def get_keys_for_dict():
    keys = (
        "pv_system_size_metadata",
        "process_output_col",
        "process_generation_and_average_cols",
        "process_efficiency_col",
        "process_metadata",
    )
    return keys



import os

from pvoutput.grid_search.grid_search import GridSearch

SHOW = True
if "CI" in os.environ:
    SHOW = False


def test_init():
    """Test that Grid search can be initiated"""
    _ = GridSearch()


def test_list_countries():
    """Get list of countries"""
    grd = GridSearch()
    countries = grd.nat_earth.list_countries()
    assert len(countries) == 258


def test_uk_grid():
    """Example 1: Get UK grid

    Use this to clip to a bounding box as well as the countries selected
    List as many countries as you want, or set to None for world-wide
    Only include search points within a certain radius of a location (see Example 3)
    Increase this if you'd like to consider systems "near" the target region (see Example 2)
    Allow some extra overlap due to inaccuracies in measuring distance
    EPSG:27700 is OSGB36 / British National Grid
    Gives a nice plot of the region and grid
    """
    grd = GridSearch()
    ukgrid = grd.generate_grid(
        bbox=[45, -15, 58, 15],
        countries=["United Kingdom"],
        radial_clip=None,
        buffer=0,
        search_radius=24.5,
        local_crs_epsg=27700,
        show=SHOW,
    )
    assert len(ukgrid) > 100


def test_luxembourg_grid():
    """Example 2: Make Luxembourg grid

    Include search radii within 50km of Luzembourgs border
    Allow some extra overlap due to inaccuracies in measuring distance
    EPSG:2169 is Luxembourg 1930 / Gauss

    """
    grd = GridSearch()
    luxgrid = grd.generate_grid(
        countries=["Luxembourg"], buffer=50, search_radius=24.5, local_crs_epsg=2169, show=SHOW
    )
    luxgrid.head()
    assert len(luxgrid) == 18


def test_sheffield_grid():
    """Make grid around Sheffield

    Only include search points within a 100km of the TUOS Physics Department
    EPSG:27700 is OSGB36 / British National Grid

    """
    grd = GridSearch()
    shefgrid = grd.generate_grid(
        radial_clip=(
            53.381,
            -1.486,
            100.0,
        ),  # Only include search points within a 100km of the TUOS Physics Department
        local_crs_epsg=27700,  # EPSG:27700 is OSGB36 / British National Grid
        show=SHOW,
    )
    assert len(shefgrid) == 29


def test_balkan_grid():
    """Plot grid around Balkan area"""
    grd = GridSearch()
    balkan_grid = grd.generate_grid(
        countries=[
            "Bosnia and Herz.",
            "Croatia",
            "Hungary",
            "Romania",
            "Bulgaria",
            "North Macedonia",
            "Kosovo",
            "Albania",
            "Montenegro",
            "Serbia",
        ],
        search_radius=24.5,
        show=SHOW,
    )
    assert len(balkan_grid) == 733



from datetime import date
from io import StringIO

import numpy as np
import pandas as pd
import pytest

from pvoutput.prcoess import process_batch_status, process_system_status


def test_process_system_status():
    pv_system_status_text = "1234;07:45,21,255,1,2;"
    one_status = process_system_status(
        pv_system_status_text=pv_system_status_text, date=date(2022, 1, 1)
    )
    assert len(one_status) == 1


def test_process_system_status_2():
    # note that the second entry has a missing data value
    pv_system_status_text = "1234;07:45,21,255,1,5;" "07:50,22,257,2;" "07:55,23,256,3,4"

    one_status = process_system_status(
        pv_system_status_text=pv_system_status_text, date=date(2022, 1, 1)
    )
    assert len(one_status) == 3
    assert (one_status["system_id"] == 1234).all()


def test_process_system_status_none():

    one_status = process_system_status(
        pv_system_status_text="no status found", date=date(2022, 1, 1)
    )
    assert len(one_status) == 0


def test_process_system_status_less_columns_two_data_points():
    # this has all missing data values
    pv_system_status_text = "1234;07:45,21,255;" "07:45,22,256"
    one_status = process_system_status(
        pv_system_status_text=pv_system_status_text, date=date(2022, 1, 1)
    )
    assert len(one_status) == 2


def test_process_system_status_less_columns_one_data_points():
    # this has all missing data values
    pv_system_status_text = "1234;07:45,21,255"
    one_status = process_system_status(
        pv_system_status_text=pv_system_status_text, date=date(2022, 1, 1)
    )
    assert len(one_status) == 1
    assert np.isnan(one_status["temperature_C"][0])


def test_process_batch_status():
    # Response text copied from
    # https://pvoutput.org/help.html#dataservice-getbatchstatus
    response_text = """
20140330;07:35,2,24;07:40,4,24;07:45,6,24;07:50,8,24;07:55,13,60;08:00,24,132
20140329;07:35,2,24;07:40,4,24;07:45,6,24;07:50,8,24;07:55,13,60;08:00,24,132
20140328;07:35,2,24;07:40,4,24;07:45,6,24;07:50,8,24;07:55,13,60;08:00,24,132"""

    correct_interpretation_csv = """
datetime,cumulative_energy_gen_Wh,instantaneous_power_gen_W,temperature_C,voltage
2014-03-28 07:35:00,2.0,24.0,,
2014-03-28 07:40:00,4.0,24.0,,
2014-03-28 07:45:00,6.0,24.0,,
2014-03-28 07:50:00,8.0,24.0,,
2014-03-28 07:55:00,13.0,60.0,,
2014-03-28 08:00:00,24.0,132.0,,
2014-03-29 07:35:00,2.0,24.0,,
2014-03-29 07:40:00,4.0,24.0,,
2014-03-29 07:45:00,6.0,24.0,,
2014-03-29 07:50:00,8.0,24.0,,
2014-03-29 07:55:00,13.0,60.0,,
2014-03-29 08:00:00,24.0,132.0,,
2014-03-30 07:35:00,2.0,24.0,,
2014-03-30 07:40:00,4.0,24.0,,
2014-03-30 07:45:00,6.0,24.0,,
2014-03-30 07:50:00,8.0,24.0,,
2014-03-30 07:55:00,13.0,60.0,,
2014-03-30 08:00:00,24.0,132.0,,"""

    df = process_batch_status(response_text)
    correct_df = pd.read_csv(
        StringIO(correct_interpretation_csv), parse_dates=["datetime"], index_col="datetime"
    )
    pd.testing.assert_frame_equal(df, correct_df)

    empty_df = process_batch_status("")
    assert empty_df.empty, "DataFrame should be empty but it was:\n{}\n".format(empty_df)

    with pytest.raises(NotImplementedError):
        process_batch_status("20140330;07:35,2,24,2,24,23.1,230.3")



"""Clipping function for coordinates"""

from typing import Iterable, Optional, Tuple, Union

import geopandas as gpd
import pandas as pd
from shapely.geometry import Point, Polygon


def clip_to_radius(
    coords: Union[pd.DataFrame, gpd.GeoDataFrame],
    latitude: float,
    longitude: float,
    radius: Optional[float] = None,
    search_radius: Optional[float] = None,
    local_crs_epsg: int = 4087,
) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
    """Clip coordinates to a radius.

    Remove any coordinates which do not lie within x.

    Args:
        coords:
            A pandas DataFrame or geopandas GeoDataFrame of coordinates with columns: latitude,
            longitude.
        latitude:
            Latitude of the center of the radial search.
        longitude:
            Longitude of the center of the radial search.
        radius:
            Set the radial search limit in km.
        search_radius:
            Optionally set the radial search limit around each grid point in kilometers. If set, the
            code will consider coords to be included if any part of the search radius overlaps the
            outter radius.
        local_crs_epsg:
            Optionally provide the EPSG code of a local Co-ordinate Reference System (CRS) for
            improved accuracy. e.g. set to 27700 (OSGB36 / British National Grid) if searching the
            British Isles.

    Returns:
        As per `coords` but containing only the subset of the input coordinates which fall within
        `radius` km of the lat/lon.
    """
    center = Point(longitude, latitude)
    radius_ = (
        gpd.GeoSeries(center)
        .set_crs("EPSG:4326")
        .to_crs(f"EPSG:{local_crs_epsg}")
        .buffer(radius * 1000.0)[0]
    )
    if search_radius is None:
        coords["selected"] = coords.within(radius_)
    else:
        coords_ = coords.to_crs(f"EPSG:{local_crs_epsg}").buffer(search_radius * 1000.0)
        coords["selected"] = coords_.intersects(radius_)
    return coords.loc[coords["selected"]].drop(columns="selected")


def bounding_box_from_radius(
    latitude: float, longitude: float, radius: float, local_crs_epsg: int = 4087
) -> Tuple[float]:
    """Convert a radial search around a given lat/lon to a bounding box.

    Args:
        latitude:
            Latitude of the center of the radial search.
        longitude:
            Longitude of the center of the radial search.
        radius:
            Set the radial search limit in km.
        local_crs_epsg:
            Optionally provide the EPSG code of a local Co-ordinate Reference System (CRS) for
            improved accuracy. e.g. set to 27700 (OSGB36 / British National Grid) if searching the
            British Isles.

    Returns:
        Four element tuple defining a bounding box: (min_lat, min_lon, max_lat, max_lon).
    """
    center = Point(longitude, latitude)
    search_radius = (
        gpd.GeoSeries(center)
        .set_crs("EPSG:4326")
        .to_crs(f"EPSG:{local_crs_epsg}")
        .buffer(radius * 1000.0)
    )
    bounds = search_radius.to_crs("EPSG:4326").bounds.loc[0].to_numpy()
    return bounds[1], bounds[0], bounds[3], bounds[2]


def buffer_bounding_box_bounds(
    bbox: Optional[Iterable[float]] = None, buffer: float = 0, local_crs_epsg: int = 4087
) -> Tuple[float]:
    """Buffer a bounding box by a distance in km.

    Args:
        bbox:
            Four element iterable defining a bounding box: [min_lat, min_lon, max_lat, max_lon].
        buffer:
            Optionally buffer the country boundaries before clipping, in kilometers.
        local_crs_epsg:
            Optionally provide the EPSG code of a local Co-ordinate Reference System (CRS) for
            improved accuracy. e.g. set to 27700 (OSGB36 / British National Grid) if searching the
            British Isles.

    Returns:
        Four element tuple defining a bounding box: (min_lat, min_lon, max_lat, max_lon).
    """
    bbox_ = buffer_bounding_box(bbox, buffer, local_crs_epsg)
    new_bounds = bbox_.to_crs("EPSG:4326").bounds.loc[0].to_numpy()
    return new_bounds[1], new_bounds[0], new_bounds[3], new_bounds[2]


def buffer_bounding_box(
    bbox: Optional[Iterable[float]] = None, buffer: float = 0, local_crs_epsg: int = 4087
) -> gpd.GeoSeries:
    """Buffer a bounding box by a distance in km.

    Args:
        bbox:
            Four element iterable defining a bounding box: [min_lat, min_lon, max_lat, max_lon].
        buffer:
            Optionally buffer the country boundaries before clipping, in kilometers.
        local_crs_epsg:
            Optionally provide the EPSG code of a local Co-ordinate Reference System (CRS) for
            improved accuracy. e.g. set to 27700 (OSGB36 / British National Grid) if searching the
            British Isles.

    Returns:
        A geopandas GeoSeries containing the buffered geometry.
    """
    bbox_ = Polygon(
        [(bbox[1], bbox[0]), (bbox[1], bbox[2]), (bbox[3], bbox[2]), (bbox[3], bbox[0])]
    )
    bbox_ = (
        gpd.GeoSeries(bbox_)
        .set_crs("EPSG:4326")
        .to_crs(f"EPSG:{local_crs_epsg}")
        .buffer(buffer * 1000.0)
    )
    return bbox_


def clip_to_bbox(
    coords: Union[pd.DataFrame, gpd.GeoDataFrame],
    bbox: Iterable[float],
    buffer: float = 0,
    search_radius: Optional[float] = None,
    local_crs_epsg: int = 4087,
) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
    """Clip coordinates to bounding box.

    Remove any coordinates which do not lie inside a bounding box.

    Args:
        coords:
            A pandas DataFrame or geopandas GeoDataFrame of coordinates with columns: latitude,
            longitude.
        bbox:
            Four element iterable defining a bounding box: [min_lat, min_lon, max_lat, max_lon].
        buffer:
            Optionally buffer the country boundaries before clipping, in kilometers.
        search_radius:
            Optionally set the radial search limit around each grid point in kilometers. If set, the
            code will consider coords to be included if any part of the search radius overlaps the
            bounding box.
        local_crs_epsg:
            Optionally provide the EPSG code of a local Co-ordinate Reference System (CRS) for
            improved accuracy. e.g. set to 27700 (OSGB36 / British National Grid) if searching the
            British Isles.

    Returns:
        As per `coords` but containing only the subset of the input coordinates which fall within
        `buffer` km of the bounding box.
    """
    if search_radius is None:
        # TODO why is bounds not used
        bounds = buffer_bounding_box_bounds(bbox, buffer, local_crs_epsg)  # noqa
        coords["selected"] = (bbox[0] <= coords.latitude <= bbox[2]) & (
            bbox[1] <= coords.longitude <= bbox[3]
        )
    else:
        bbox_ = buffer_bounding_box(bbox, buffer, local_crs_epsg)[0]
        coords_ = coords.to_crs(f"EPSG:{local_crs_epsg}").buffer(search_radius * 1000.0)
        coords["selected"] = coords_.intersects(bbox_)
    return coords.loc[coords["selected"]].drop(columns="selected")


def clip_to_countries(
    coords: gpd.GeoDataFrame,
    world: gpd.GeoDataFrame,
    countries: Iterable[str],
    buffer: float = 0,
    search_radius: Optional[float] = None,
) -> gpd.GeoDataFrame:
    """Clip coordinates to country boundaries.

    Given a set of coordinates, some country boundary definitions, a list of countries, and a buffer
    distance, return the coords which fall within `buffer` kilometers of the listed countries'
    boundaries.

    Args:
        coords:
            A geopandas GeoDataFrame containing latitudes, longitudes and geometries for a set of
            coordinates.
        world:
            A geopandas GeoDataFrame of world boundaries geomteries, as returned by
            `get_world_boundaries()`.
        countries:
            A list of country names to clip the coords to.
        buffer:
            Optionally buffer the country boundaries before clipping, in kilometers.
        search_radius:
            Optionally set the radial search limit around each grid point in kilometers. If set, the
            code will consider coords to be included if any part of the search radius overlaps the
            country.

    Returns:
        As per `coords` but containing only the subset of the input coordinates which fall within
        `buffer` km of the given countries.
    """
    countries_ = world[world.name.isin(countries)]
    countries_ = countries_.dissolve().buffer(buffer * 1000.0)[0]
    if search_radius is None:
        coords["selected"] = coords.within(countries_)
    else:
        # Consider points outside the selected region whose search radius overlaps the region
        coords_ = coords.buffer(search_radius * 1000.0)
        coords["selected"] = coords_.intersects(countries_)
    return coords.loc[coords["selected"]].drop(columns="selected")



"""Main PV Output class to get data from pvoutput.org"""

import logging
import os
import time
import warnings
from datetime import date, datetime, timedelta
from io import StringIO
from typing import Dict, Iterable, List, Optional, Union
from urllib.parse import urljoin

import numpy as np
import pandas as pd
import requests
import tables

from pvoutput.consts import (
    BASE_URL,
    CONFIG_FILENAME,
    ONE_DAY,
    PV_OUTPUT_DATE_FORMAT,
    RATE_LIMIT_PARAMS_TO_API_HEADERS,
)
from pvoutput.daterange import DateRange, merge_date_ranges_to_years
from pvoutput.exceptions import NoStatusFound, RateLimitExceeded
from pvoutput.prcoess import process_batch_status, process_system_status
from pvoutput.utils import (
    _get_param_from_config_file,
    _get_response,
    _print_and_log,
    get_date_ranges_to_download,
    sort_and_de_dupe_pv_system,
    system_id_to_hdf_key,
)

_LOG = logging.getLogger("pvoutput")


class PVOutput:
    """
    Main PV Output class

    Attributes:
        api_key
        system_id
        rate_limit_remaining
        rate_limit_total
        rate_limit_reset_time
        data_service_url
    """

    def __init__(
        self,
        api_key: str = os.getenv("API_KEY"),
        system_id: str = os.getenv("SYSTEM_ID"),
        config_filename: Optional[str] = CONFIG_FILENAME,
        data_service_url: Optional[str] = os.getenv("DATA_SERVICE_URL"),
    ):
        """
        Init

        Args:
            api_key: Your API key from PVOutput.org.
            system_id: Your system ID from PVOutput.org.  If you don't have a
                PV system then you can register with PVOutput.org and select
                the 'energy consumption only' box.
            config_filename: Optional, the filename of the .yml config file.
            data_service_url: Optional.  If you have subscribed to
                PVOutput.org's data service then add the data service URL here.
                This string must end in '.org'.
        """
        self.api_key = api_key
        self.system_id = system_id
        self.rate_limit_remaining = None
        self.rate_limit_total = None
        self.rate_limit_reset_time = None
        self.data_service_url = data_service_url

        # Set from config file if None
        for param_name in ["api_key", "system_id"]:
            if getattr(self, param_name) is None:
                try:
                    param_value_from_config = _get_param_from_config_file(
                        param_name, config_filename
                    )
                except Exception as e:
                    msg = (
                        "Error loading configuration parameter {param_name}"
                        " from config file {filename}.  Either pass"
                        " {param_name} into PVOutput constructor, or create"
                        " config file {filename}.  {exception}".format(
                            param_name=param_name, filename=CONFIG_FILENAME, exception=e
                        )
                    )
                    print(msg)
                    _LOG.exception(msg)
                    raise
                setattr(self, param_name, param_value_from_config)
            # Convert to strings
            setattr(self, param_name, str(getattr(self, param_name)))

        # Check for data_service_url
        if self.data_service_url is None:
            try:
                self.data_service_url = _get_param_from_config_file(
                    "data_service_url", config_filename
                )
            except KeyError:
                pass
            except FileNotFoundError:
                pass

        if self.data_service_url is not None:
            if not self.data_service_url.strip("/").endswith(".org"):
                raise ValueError("data_service_url must end in '.org'")

    def search(
        self,
        query: str,
        lat: Optional[float] = None,
        lon: Optional[float] = None,
        include_country: bool = True,
        **kwargs,
    ) -> pd.DataFrame:
        """Search for PV systems.

        Some quirks of the PVOutput.org API:
            - The maximum number of results returned by PVOutput.org is 30.
                If the number of returned results is 30, then there is no
                indication of whether there are exactly 30 search results,
                or if there are more than 30.  Also, there is no way to
                request additional 'pages' of search results.
            - The maximum search radius is 25km

        Args:
            query: string, see https://pvoutput.org/help.html#search
                e.g. '5km'.
            lat: float, e.g. 52.0668589
            lon: float, e.g. -1.3484038
            include_country: bool, whether or not to include the country name
                with the returned postcode.

        Returns:
            pd.DataFrame, one row per search results.  Index is PV system ID.
                Columns:
                    name,
                    system_DC_capacity_W,
                    address,  # If `include_country` is True then address is
                              # 'country> <postcode>',
                              # else address is '<postcode>'.
                    orientation,
                    num_outputs,
                    last_output,
                    panel,
                    inverter,
                    distance_km,
                    latitude,
                    longitude
        """
        api_params = {"q": query, "country": int(include_country)}

        if lat is not None and lon is not None:
            api_params["ll"] = "{:f},{:f}".format(lat, lon)

        pv_systems_text = self._api_query(service="search", api_params=api_params, **kwargs)

        pv_systems = pd.read_csv(
            StringIO(pv_systems_text),
            names=[
                "name",
                "system_DC_capacity_W",
                "address",
                "orientation",
                "num_outputs",
                "last_output",
                "system_id",
                "panel",
                "inverter",
                "distance_km",
                "latitude",
                "longitude",
            ],
            index_col="system_id",
        )

        return pv_systems

    def get_status(
        self,
        pv_system_id: int,
        date: Union[str, datetime],
        historic: bool = True,
        timezone: Optional[str] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Get PV system status (e.g. power generation) for one day.

        The returned DataFrame will be empty if the PVOutput API
        returns 'status 400: No status found'.

        Args:
            pv_system_id: int
            date: str in format YYYYMMDD; or datetime
                (localtime of the PV system)
            timezone: the timezone of the systems. This will be used to add to the datetime.
                If None, it is not added

        Returns:
            pd.DataFrame:
                index: datetime (DatetimeIndex, localtime of the PV system)
                columns:  (all np.float64):
                    cumulative_energy_gen_Wh,
                    energy_efficiency_kWh_per_kW,
                    instantaneous_power_gen_W,
                    average_power_gen_W,
                    power_gen_normalised,
                    energy_consumption_Wh,
                    power_demand_W,
                    temperature_C,
                    voltage
        """
        _LOG.info("system_id %d: Requesting system status for %s", pv_system_id, date)
        date = date_to_pvoutput_str(date)
        _check_date(date)

        api_params = {
            "d": date,  # date, YYYYMMDD, localtime of the PV system
            "h": int(historic is True),  # We want historical data.
            "limit": 288,  # API limit is 288 (num of 5-min periods per day).
            "ext": 0,  # Extended data; we don't want extended data.
            "sid1": pv_system_id,  # SystemID.
        }

        try:
            pv_system_status_text = self._api_query(
                service="getstatus", api_params=api_params, **kwargs
            )
        except NoStatusFound:
            _LOG.info("system_id %d: No status found for date %s", pv_system_id, date)
            pv_system_status_text = ""

        # See https://pvoutput.org/help.html#api-getstatus but make sure
        # you read the 'History Query' subsection, as a historical query
        # has slightly different return columns compared to a non-historical
        # query!
        columns = (
            [
                "cumulative_energy_gen_Wh",
                "energy_efficiency_kWh_per_kW",
                "instantaneous_power_gen_W",
                "average_power_gen_W",
                "power_gen_normalised",
                "energy_consumption_Wh",
                "power_demand_W",
                "temperature_C",
                "voltage",
            ]
            if historic
            else [
                "cumulative_energy_gen_Wh",
                "instantaneous_power_gen_W",
                "energy_consumption_Wh",
                "power_demand_W",
                "power_gen_normalised",
                "temperature_C",
                "voltage",
            ]
        )

        pv_system_status = pd.read_csv(
            StringIO(pv_system_status_text),
            lineterminator=";",
            names=["date", "time"] + columns,
            parse_dates={"datetime": ["date", "time"]},
            index_col=["datetime"],
            dtype={col: np.float64 for col in columns},
        ).sort_index()

        # add timezone
        if timezone is not None:
            pv_system_status = pv_system_status.tz_localize(timezone).tz_convert("UTC")

        return pv_system_status

    def get_system_status(
        self,
        pv_system_ids: List[int],
        date: Union[str, datetime],
        timezone: Optional[str] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Get Batch of PV system status (e.g. power generation) for one day, for multiple systems

        The returned DataFrame will be empty if the PVOutput API
        returns 'status 400: No status found'.

        Args:
            pv_system_ids: list of ints.
                If you have a subscription service then multiple (up to 50)
                pv systems status can be queries at once
            date: str in format YYYYMMDD; or datetime
                (localtime of the PV system)
            timezone: the timezone of the systems. This will be used to add to the datetime.
                If None, it is not added

        Returns:
            pd.DataFrame:
                columns:  (all np.float64):
                    system_id,
                    datetime,
                    instantaneous_power_gen_W,
                    cumulative_energy_gen_Wh,
                    instantaneous_power_gen_W,
                    energy_consumption_Wh",
                    temperature_C,
                    voltage,
        """
        _LOG.info(f"system_ids {pv_system_ids}: Requesting batch system status for %s", date)
        date = date_to_pvoutput_str(date)
        _check_date(date)

        # join the system ids with a column
        all_pv_system_id = ",".join([str(idx) for idx in pv_system_ids])

        api_params = {
            "dt": date,  # date, YYYYMMDD, localtime of the PV system
            "sid1": all_pv_system_id,  # SystemID.
        }

        try:
            pv_system_status_text = self._api_query(
                service="getsystemstatus", api_params=api_params, **kwargs
            )

        except NoStatusFound:
            _LOG.info(f"system_id {all_pv_system_id}: No status found for date %s", date)
            pv_system_status_text = "no status found"

        # each pv system is on a new line
        pv_systems_status_text = pv_system_status_text.split("\n")

        pv_system_status = []
        for pv_system_status_text in pv_systems_status_text:

            try:
                one_pv_system_status = process_system_status(
                    pv_system_status_text=pv_system_status_text, date=date
                )
            except Exception as e:
                _LOG.error(
                    f"Could not change raw text into dataframe. Raw text is {pv_system_status_text}"
                )
                raise e

            pv_system_status.append(one_pv_system_status)

        pv_system_status = pd.concat(pv_system_status)
        pv_system_status.reset_index(inplace=True)

        # add timezone
        if timezone is not None:
            pv_system_status["datetime"] = (
                pd.DatetimeIndex(pv_system_status["datetime"])
                .tz_localize(timezone)
                .tz_convert("UTC")
            )

        return pv_system_status

    def get_batch_status(
        self,
        pv_system_id: int,
        date_to: Optional[Union[str, datetime]] = None,
        max_retries: Optional[int] = 1000,
        **kwargs,
    ) -> Union[None, pd.DataFrame]:
        """Get batch PV system status (e.g. power generation).

        The returned DataFrame will be empty if the PVOutput API
        returns 'status 400: No status found'.

        Data returned is limited to the last 366 days per request.
        To retrieve older data, use the date_to parameter.

        The PVOutput getbatchstatus API is asynchronous.  When it's first
        called, it replies to say 'accepted'.  This function will then
        wait a minute and call the API again to see if the data is ready.
        Set `max_retries` to 1 if you want to return immediately, even
        if data isn't ready yet (and hence this function will return None)

        https://pvoutput.org/help.html#dataservice-getbatchstatus

        Args:
            pv_system_id: int
            date_to: str in format YYYYMMDD; or datetime
                (localtime of the PV system).  The returned timeseries will
                include 366 days of data: from YYYY-1MMDD to YYYYMMDD inclusive
            max_retries: int, number of times to retry after receiving
                a '202 Accepted' request.  Set `max_retries` to 1 if you want
                to return immediately, even if data isn't ready yet (and hence
                this function will return None).

        Returns:
            None (if data isn't ready after retrying max_retries times) or
            pd.DataFrame:
                index: datetime (DatetimeIndex, localtime of the PV system)
                columns:  (all np.float64):
                    cumulative_energy_gen_Wh,
                    instantaneous_power_gen_W,
                    temperature_C,
                    voltage
        """
        api_params = {"sid1": pv_system_id}

        _set_date_param(date_to, api_params, "dt")

        for retry in range(max_retries):
            try:
                pv_system_status_text = self._api_query(
                    service="getbatchstatus", api_params=api_params, use_data_service=True, **kwargs
                )
            except NoStatusFound:
                _LOG.info("system_id %d: No status found for date_to %s", pv_system_id, date_to)
                pv_system_status_text = ""
                break

            if "Accepted 202" in pv_system_status_text:
                if retry == 0:
                    _print_and_log("Request accepted.")
                if retry < max_retries - 1:
                    _print_and_log("Sleeping for 1 second.")
                    time.sleep(1)
                else:
                    _print_and_log(
                        "Call get_batch_status again in a minute to see if" " results are ready."
                    )
            else:
                break
        else:
            return

        return process_batch_status(pv_system_status_text)

    def get_metadata(self, pv_system_id: int, **kwargs) -> pd.Series:
        """Get metadata for a single PV system.

        Args:
            pv_system_id: int

        Returns:
            pd.Series.  Index is:
                name,
                system_DC_capacity_W,
                address,
                num_panels,
                panel_capacity_W_each,
                panel_brand,
                num_inverters,
                inverter_capacity_W,
                inverter_brand,
                orientation,
                array_tilt_degrees,
                shade,
                install_date,
                latitude,
                longitude,
                status_interval_minutes,
                secondary_num_panels,
                secondary_panel_capacity_W_each,
                secondary_orientation,
                secondary_array_tilt_degrees
        """
        pv_metadata_text = self._api_query(
            service="getsystem",
            api_params={
                "array2": 1,  # Provide data about secondary array, if present.
                "tariffs": 0,
                "teams": 0,
                "est": 0,
                "donations": 0,
                "sid1": pv_system_id,  # SystemID
                "ext": 0,  # Include extended data?
            },
            **kwargs,
        )

        _LOG.debug(f"getting metadata for {pv_system_id}")

        pv_metadata = pd.read_csv(
            StringIO(pv_metadata_text),
            lineterminator=";",
            names=[
                "name",
                "system_DC_capacity_W",
                "address",
                "num_panels",
                "panel_capacity_W_each",
                "panel_brand",
                "num_inverters",
                "inverter_capacity_W",
                "inverter_brand",
                "orientation",
                "array_tilt_degrees",
                "shade",
                "install_date",
                "latitude",
                "longitude",
                "status_interval_minutes",
                "secondary_num_panels",
                "secondary_panel_capacity_W_each",
                "secondary_orientation",
                "secondary_array_tilt_degrees",
            ],
            parse_dates=["install_date"],
            nrows=1,
        ).squeeze()
        pv_metadata["system_id"] = pv_system_id
        pv_metadata.name = pv_system_id
        return pv_metadata

    def get_statistic(
        self,
        pv_system_id: int,
        date_from: Optional[Union[str, date]] = None,
        date_to: Optional[Union[str, date]] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Get summary stats for a single PV system.

        Args:
            pv_system_id: int
            date_from
            date_to

        Returns:
            pd.DataFrame:
                total_energy_gen_Wh,
                energy_exported_Wh,
                average_daily_energy_gen_Wh,
                minimum_daily_energy_gen_Wh,
                maximum_daily_energy_gen_Wh,
                average_efficiency_kWh_per_kW,
                num_outputs,  # The number of days for which there's >= 1 val.
                actual_date_from,
                actual_date_to,
                record_efficiency_kWh_per_kW,
                record_efficiency_date,
                query_date_from,
                query_date_to
        """
        if date_from and not date_to:
            date_to = pd.Timestamp.now().date()
        if date_to and not date_from:
            date_from = pd.Timestamp("1900-01-01").date()

        api_params = {
            "c": 0,  # consumption and import
            "crdr": 0,  # credits / debits
            "sid1": pv_system_id,  # SystemID
        }

        _set_date_param(date_from, api_params, "df")
        _set_date_param(date_to, api_params, "dt")

        try:
            pv_metadata_text = self._api_query(
                service="getstatistic", api_params=api_params, **kwargs
            )
        except NoStatusFound:
            pv_metadata_text = ""

        columns = [
            "total_energy_gen_Wh",
            "energy_exported_Wh",
            "average_daily_energy_gen_Wh",
            "minimum_daily_energy_gen_Wh",
            "maximum_daily_energy_gen_Wh",
            "average_efficiency_kWh_per_kW",
            "num_outputs",
            "actual_date_from",
            "actual_date_to",
            "record_efficiency_kWh_per_kW",
            "record_efficiency_date",
        ]
        date_cols = ["actual_date_from", "actual_date_to", "record_efficiency_date"]
        numeric_cols = set(columns) - set(date_cols)
        pv_metadata = pd.read_csv(
            StringIO(pv_metadata_text),
            names=columns,
            dtype={col: np.float32 for col in numeric_cols},
            parse_dates=date_cols,
        )
        if pv_metadata.empty:
            data = {col: np.float32(np.NaN) for col in numeric_cols}
            data.update({col: pd.NaT for col in date_cols})
            pv_metadata = pd.DataFrame(data, index=[pv_system_id])
        else:
            pv_metadata.index = [pv_system_id]

        pv_metadata["query_date_from"] = pd.Timestamp(date_from) if date_from else pd.NaT
        pv_metadata["query_date_to"] = pd.Timestamp(date_to) if date_to else pd.Timestamp.now()
        return pv_metadata

    def _get_statistic_with_cache(
        self,
        store_filename: str,
        pv_system_id: int,
        date_from: Optional[Union[str, date]] = None,
        date_to: Optional[Union[str, date]] = None,
        **kwargs,
    ) -> pd.Series:
        """
        Get Statistic using cache

        Will try to get stats from store_filename['statistics'].  If this
        fails, or if date_to > query_date_to, or if
        date_from < query_date_from, then will call the API.  Note that the aim
        of this function is just to find the relevant actual_date_from and
        actual_date_to, so this function does not respect the other params.

        Args:
            store_filename: cache filenamte
            pv_system_id: pv system id
            date_from: the start date we want statistics from
            date_to: the end date we want statistics from
            **kwargs:

        Returns: Pandas data series holding various statistics

        """
        if date_from:
            date_from = pd.Timestamp(date_from).date()
        if date_to:
            date_to = pd.Timestamp(date_to).date()

        def _get_fresh_statistic():
            _LOG.info("pv_system %d: Getting fresh statistic.", pv_system_id)
            stats = self.get_statistic(pv_system_id, **kwargs)
            with pd.HDFStore(store_filename, mode="a") as store:
                try:
                    store.remove(key="statistics", where="index=pv_system_id")
                except KeyError:
                    pass
                store.append(key="statistics", value=stats)
            return stats

        try:
            stats = pd.read_hdf(store_filename, key="statistics", where="index=pv_system_id")
        except (FileNotFoundError, KeyError):
            return _get_fresh_statistic()

        if stats.empty:
            return _get_fresh_statistic()

        query_date_from = stats.iloc[0]["query_date_from"]
        query_date_to = stats.iloc[0]["query_date_to"]

        if (
            not pd.isnull(date_from)
            and not pd.isnull(query_date_from)
            and date_from < query_date_from.date()
        ):
            return _get_fresh_statistic()

        if not pd.isnull(date_to) and date_to > query_date_to.date():
            return _get_fresh_statistic()

        return stats

    def download_multiple_systems_to_disk(
        self,
        system_ids: Iterable[int],
        start_date: datetime,
        end_date: datetime,
        output_filename: str,
        timezone: Optional[str] = None,
        min_data_availability: Optional[float] = 0.5,
        use_get_batch_status_if_available: Optional[bool] = True,
    ):
        """Download multiple PV system IDs to disk.

        Data is saved to `output_filename` in HDF5 format.  The exact data
        format is documented in
        https://github.com/openclimatefix/pvoutput/blob/master/docs/dataset.md

        This function is designed to be run for days (!) downloading
        gigabytes of PV data :)  As such, this function can be safely
        interrupted and re-started.  All the state required to re-start
        is stored in the HDF5 file.

        Add appropriate handlers the Python logger `pvoutput` to see progress.

        Args:
            system_ids: List of PV system IDs to download.
            start_date: Start of date range to download.
            end_date: End of date range to download.
            output_filename: HDF5 filename to write data to.
            timezone: String representation of timezone of timeseries data.
                e.g. 'Europe/London'.
            min_data_availability: A float in the range [0, 1].  1 means only
                accept PV systems which have no days of missing data.  0 means
                accept all PV systems, no matter if they have missing data.
                Note that the data availability is measured against the date
                range for which the PV system has data available, not from
                the date range passed into this function.
            use_get_batch_status_if_available: Bool.  If true then will use
                PVOutput's getbatchstatus API (which must be paid for, and
                `data_service_url` must be set in `~/.pvoutput.yml` or when
                initialising the PVOutput object).
        """
        n = len(system_ids)
        for i, pv_system_id in enumerate(system_ids):
            _LOG.info("**********************")
            msg = "system_id {:d}: {:d} of {:d} ({:%})".format(pv_system_id, i + 1, n, (i + 1) / n)
            _LOG.info(msg)
            print("\r", msg, end="", flush=True)

            # Sorted list of DateRange objects.  For each DateRange,
            # we need to download from start_date to end_date inclusive.
            date_ranges_to_download = get_date_ranges_to_download(
                output_filename, pv_system_id, start_date, end_date
            )

            # How much data is actually available?
            date_ranges_to_download = self._filter_date_range(
                output_filename, pv_system_id, date_ranges_to_download, min_data_availability
            )

            if not date_ranges_to_download:
                _LOG.info("system_id %d: No data left to download :)", pv_system_id)
                continue

            _LOG.info(
                "system_id %d: Will download these date ranges: %s",
                pv_system_id,
                date_ranges_to_download,
            )

            if use_get_batch_status_if_available:
                if self.data_service_url:
                    self._download_multiple_using_get_batch_status(
                        output_filename, pv_system_id, date_ranges_to_download, timezone
                    )
                else:
                    raise ValueError("data_service_url is not set!")
            else:
                self._download_multiple_using_get_status(
                    output_filename, pv_system_id, date_ranges_to_download, timezone
                )

    def get_insolation_forecast(
        self,
        date: Union[str, datetime],
        pv_system_id: Optional[int] = None,
        timezone: Optional[str] = None,
        lat: Optional[float] = None,
        lon: Optional[float] = None,
        **kwargs,
    ):
        """Get Insolation forecast data

        This is for a given site, or a given location defined by
        longitude and latitude.

        This is the estimated output for the site
        based on ideal weather conditions. Also factors in site age, reducing
        ouput by 1% each year, shade and orientation. Need donation mode enabled.
        See https://pvoutput.org/help.html#api-getinsolation

        Args:
           date: str in format YYYYMMDD; or datetime
            (localtime of the PV system)
            pv_system_id: int
            timezone: str
            lat: float e.g. -27.4676
            lon: float e.g. 153.0279
            **kwargs:


        Returns: dataframe of the insolution forecast

        """
        date = date_to_pvoutput_str(date)
        _check_date(date, prediction=True)
        api_params = {
            "d": date,  # date, YYYYMMDD, localtime of the PV system
            "sid1": pv_system_id,  # SystemID.
            "tz": timezone,  # defaults to configured timezone of system otherwise GMT
        }
        if lat is not None and lon is not None:
            api_params["ll"] = "{:f},{:f}".format(lat, lon)

        try:
            pv_insolation_text = self._api_query(
                service="getinsolation", api_params=api_params, **kwargs
            )
        except NoStatusFound:
            _LOG.info("system_id %d: No status found for date %s", pv_system_id, date)
            pv_insolation_text = ""

        columns = ["predicted_power_gen_W", "predicted_cumulative_energy_gen_Wh"]
        pv_insolation = pd.read_csv(
            StringIO(pv_insolation_text),
            lineterminator=";",
            names=["time"] + columns,
            dtype={col: np.float64 for col in columns},
        ).sort_index()
        pv_insolation.index = pd.to_datetime(
            date + " " + pv_insolation.time, format="%Y-%m-%d %H:%M"
        )
        pv_insolation.drop("time", axis=1, inplace=True)
        return pv_insolation

    def _filter_date_range(
        self,
        store_filename: str,
        system_id: int,
        date_ranges: Iterable[DateRange],
        min_data_availability: Optional[float] = 0.5,
    ) -> List[DateRange]:
        """Check getstatistic to see if system_id has data for all date ranges.

        Args:
            system_id: PV system ID.
            store_filename: HDF5 filename to cache statistics to / from.
            date_ranges: List of DateRange objects.
            min_data_availability: A float in the range [0, 1].  1 means only
                accept PV systems which have no days of missing data.  0 means
                accept all PV systems, no matter if they have missing data.
        """
        if not date_ranges:
            return date_ranges

        stats = self._get_statistic_with_cache(
            store_filename,
            system_id,
            date_to=date_ranges[-1].end_date,
            wait_if_rate_limit_exceeded=True,
        ).squeeze()

        if pd.isnull(stats["actual_date_from"]) or pd.isnull(stats["actual_date_to"]):
            _LOG.info("system_id %d: Stats say there is no data!", system_id)
            return []

        timeseries_date_range = DateRange(stats["actual_date_from"], stats["actual_date_to"])

        data_availability = stats["num_outputs"] / (timeseries_date_range.total_days() + 1)

        if data_availability < min_data_availability:
            _LOG.info(
                "system_id %d: Data availability too low!  Only %.0f %%.",
                system_id,
                data_availability * 100,
            )
            return []

        new_date_ranges = []
        for date_range in date_ranges:
            new_date_range = date_range.intersection(timeseries_date_range)
            if new_date_range:
                new_date_ranges.append(new_date_range)
        return new_date_ranges

    def _download_multiple_using_get_batch_status(
        self, output_filename, pv_system_id, date_ranges_to_download, timezone: Optional[str] = None
    ):
        years = merge_date_ranges_to_years(date_ranges_to_download)
        dates_to = [year.end_date for year in years]
        total_rows = self._download_multiple_worker(
            output_filename, pv_system_id, dates_to, timezone, use_get_status=False
        )

        # Re-load data, sort, remove duplicate indicies, append back
        if total_rows:
            with pd.HDFStore(output_filename, mode="a", complevel=9) as store:
                sort_and_de_dupe_pv_system(store, pv_system_id)

    def _download_multiple_using_get_status(
        self, output_filename, pv_system_id, date_ranges_to_download, timezone: Optional[str] = None
    ):
        for date_range in date_ranges_to_download:
            dates = date_range.date_range()
            self._download_multiple_worker(
                output_filename, pv_system_id, dates, timezone, use_get_status=True
            )

    def _download_multiple_worker(
        self, output_filename, pv_system_id, dates, timezone, use_get_status
    ) -> int:
        """
        Download data with multiple workers

        Returns:
            total number of rows downloaded
        """
        total_rows = 0
        for date_to_load in dates:
            _LOG.info("system_id %d: Requesting date: %s", pv_system_id, date_to_load)
            datetime_of_api_request = pd.Timestamp.utcnow()
            if use_get_status:
                timeseries = self.get_status(
                    pv_system_id, date_to_load, wait_if_rate_limit_exceeded=True
                )
            else:
                timeseries = self.get_batch_status(pv_system_id, date_to=date_to_load)
            if timeseries.empty:
                _LOG.info(
                    "system_id %d: Got empty timeseries back for %s", pv_system_id, date_to_load
                )
                if use_get_status:
                    _append_missing_date_range(
                        output_filename,
                        pv_system_id,
                        date_to_load,
                        date_to_load,
                        datetime_of_api_request,
                    )
                else:
                    _append_missing_date_range(
                        output_filename,
                        pv_system_id,
                        date_to_load - timedelta(days=365),
                        date_to_load,
                        datetime_of_api_request,
                    )
            else:
                total_rows += len(timeseries)
                _LOG.info(f"Adding timezone {timezone} to {total_rows} rows")
                timeseries = timeseries.tz_localize(timezone)
                _LOG.info(
                    "system_id: %d: %d rows retrieved: %s to %s",
                    pv_system_id,
                    len(timeseries),
                    timeseries.index[0],
                    timeseries.index[-1],
                )
                if use_get_status:
                    check_pv_system_status(timeseries, date_to_load)
                else:
                    _record_gaps(
                        output_filename,
                        pv_system_id,
                        date_to_load,
                        timeseries,
                        datetime_of_api_request,
                    )
                timeseries["datetime_of_API_request"] = datetime_of_api_request
                timeseries["query_date"] = pd.Timestamp(date_to_load)
                key = system_id_to_hdf_key(pv_system_id)
                with pd.HDFStore(output_filename, mode="a", complevel=9) as store:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", tables.NaturalNameWarning)
                        store.append(key=key, value=timeseries, data_columns=True)

        _LOG.info("system_id %d: %d total rows downloaded", pv_system_id, total_rows)
        return total_rows

    def _api_query(
        self,
        service: str,
        api_params: Dict,
        wait_if_rate_limit_exceeded: bool = False,
        use_data_service: bool = False,
    ) -> str:
        """Send API request to PVOutput.org and return content text.

        Args:
            service: string, e.g. 'search' or 'getstatus'
            api_params: dict
            wait_if_rate_limit_exceeded: bool
            use_data_service: bool

        Raises:
            NoStatusFound
            RateLimitExceeded
        """
        get_response_func = (
            self._get_data_service_response if use_data_service else self._get_api_response
        )

        try:
            response = get_response_func(service, api_params)
        except Exception as e:
            _LOG.exception(e)
            raise

        try:
            return self._process_api_response(response)
        except RateLimitExceeded:
            msg = "PVOutput.org API rate limit exceeded!" "  Rate limit will be reset at {}".format(
                self.rate_limit_reset_time
            )
            _print_and_log(msg)
            if wait_if_rate_limit_exceeded:
                self.wait_for_rate_limit_reset()
                return self._api_query(service, api_params, wait_if_rate_limit_exceeded=False)

            raise RateLimitExceeded(response, msg)

    def _get_api_response(self, service: str, api_params: Dict) -> requests.Response:
        """
        Get the non-data service (free) response from pvoutput.org

        Args:
            service: string, e.g. 'search', 'getstatus'
            api_params: dict
        """
        self._check_api_params()
        # Create request headers
        headers = {
            "X-Rate-Limit": "1",
            "X-Pvoutput-Apikey": self.api_key,
            "X-Pvoutput-SystemId": self.system_id,
        }

        api_url = urljoin(BASE_URL, "service/r2/{}.jsp".format(service))

        return _get_response(api_url, api_params, headers)

    def _get_data_service_response(self, service: str, api_params: Dict) -> requests.Response:
        """
        Get the data service response from pvoutput.org

        Args:
            service: string, e.g. 'getbatchstatus'
            api_params: dict
        """
        self._check_api_params()
        if self.data_service_url is None:
            raise ValueError("data_service_url must be set to use the data service!")

        headers = {"X-Rate-Limit": "1"}
        api_params = api_params.copy()
        api_params["key"] = self.api_key
        api_params["sid"] = self.system_id

        api_url = urljoin(self.data_service_url, "data/r2/{}.jsp".format(service))

        return _get_response(api_url, api_params, headers)

    def _check_api_params(self):
        # Check we have relevant login details:
        for param_name in ["api_key", "system_id"]:
            if getattr(self, param_name) is None:
                raise ValueError("Please set the {} parameter.".format(param_name))

    def _set_rate_limit_params(self, headers):
        for param_name, header_key in RATE_LIMIT_PARAMS_TO_API_HEADERS.items():
            header_value = int(headers[header_key])
            setattr(self, param_name, header_value)

        self.rate_limit_reset_time = pd.Timestamp.utcfromtimestamp(self.rate_limit_reset_time)
        self.rate_limit_reset_time = self.rate_limit_reset_time.tz_localize("utc")

        _LOG.debug("%s", self.rate_limit_info())

    def rate_limit_info(self) -> Dict:
        """Get the rate limit information"""
        info = {}
        for param_name in RATE_LIMIT_PARAMS_TO_API_HEADERS:
            info[param_name] = getattr(self, param_name)
        return info

    def _process_api_response(self, response: requests.Response) -> str:
        """Turns an API response into text.

        Args:
            response: from _get_api_response()

        Returns:
            content of the response.

        Raises:
            UnicodeDecodeError
            NoStatusFound
            RateLimitExceeded
        """
        if response.status_code == 400:
            raise NoStatusFound(response=response)

        if response.status_code != 403:
            try:
                response.raise_for_status()
            except Exception as e:
                msg = "Bad status code! Response content = {}. Exception = {}".format(
                    response.content, e
                )
                _LOG.exception(msg)
                raise e.__class__(msg)

        self._set_rate_limit_params(response.headers)

        # Did we overshoot our quota?
        if response.status_code == 403 and self.rate_limit_remaining <= 0:
            raise RateLimitExceeded(response=response)

        try:
            content = response.content.decode("latin1").strip()
        except Exception as e:
            msg = "Error decoding this string: {}\n{}".format(response.content, e)
            _LOG.exception(msg)
            raise

        # If we get to here then the content is valid :)
        return content

    def wait_for_rate_limit_reset(self, do_sleeping: bool = True) -> int:
        """
        Wait for reset limit

        Args:
            do_sleeping: bool to do the sleeping, or not.

        Returns: The number of seconds needed to sleep
        """
        utc_now = pd.Timestamp.utcnow()
        timedelta_to_wait = self.rate_limit_reset_time - utc_now
        timedelta_to_wait += timedelta(minutes=3)  # Just for safety
        secs_to_wait = timedelta_to_wait.total_seconds()
        retry_time_utc = utc_now + timedelta_to_wait

        # good to have the retry time in local so that user see 'their' time
        # retry_time_local = retry_time_utc.tz_convert(tz=datetime.now(tzlocal()).tzname())
        retry_time_local = retry_time_utc
        _print_and_log(
            "Waiting {:.0f} seconds.  Will retry at {} UTC".format(secs_to_wait, retry_time_local)
        )
        if do_sleeping:
            time.sleep(secs_to_wait)

        return secs_to_wait


def date_to_pvoutput_str(date: Union[str, datetime]) -> str:
    """Convert datetime to date string for PVOutput.org in YYYYMMDD format."""
    if isinstance(date, str):
        try:
            datetime.strptime(date, PV_OUTPUT_DATE_FORMAT)
        except ValueError:
            return pd.Timestamp(date).strftime(PV_OUTPUT_DATE_FORMAT)
        else:
            return date
    return date.strftime(PV_OUTPUT_DATE_FORMAT)


def _check_date(date: str, prediction=False):
    """Check that date string

    1. conforms to YYYYMMDD format,
    2. that the date isn't in the future.

    Raises:
        ValueError if the date is 'bad'.
    """
    dt = datetime.strptime(date, PV_OUTPUT_DATE_FORMAT)
    if dt > datetime.now() and not prediction:
        raise ValueError(
            ""
            "date should not be in the future.  Got {}.  Current date is {}.".format(
                date, datetime.now()
            )
        )


def _set_date_param(dt, api_params, key):
    if dt is not None:
        dt = date_to_pvoutput_str(dt)
        _check_date(dt)
        api_params[key] = dt


def check_pv_system_status(pv_system_status: pd.DataFrame, requested_date: date):
    """Checks the DataFrame returned by get_pv_system_status.

    Args:
        pv_system_status: DataFrame returned by get_pv_system_status
        requested_date: date.

    Raises:
        ValueError if the DataFrame is incorrect.
    """
    if not isinstance(pv_system_status, pd.DataFrame):
        raise ValueError("pv_system_status must be a dataframe")
    if not pv_system_status.empty:
        index = pv_system_status.index
        for d in [index[0], index[-1]]:
            if not requested_date <= d.date() <= requested_date + ONE_DAY:
                raise ValueError(
                    "A date in the index is outside the expected range."
                    " Date from index={}, requested_date={}".format(d, requested_date)
                )


def _append_missing_date_range(
    output_filename, pv_system_id, missing_start_date, missing_end_date, datetime_of_api_request
):

    data = {
        "missing_start_date_PV_localtime": pd.Timestamp(missing_start_date),
        "missing_end_date_PV_localtime": pd.Timestamp(missing_end_date),
        "datetime_of_API_request": datetime_of_api_request,
    }
    new_missing_date_range = pd.DataFrame(data, index=[pv_system_id])
    new_missing_date_range.index.name = "pv_system_id"
    _LOG.info(
        "system_id %d: Recording missing date range from %s to %s",
        pv_system_id,
        missing_start_date,
        missing_end_date,
    )
    with pd.HDFStore(output_filename, mode="a", complevel=9) as store:
        store.append(key="missing_dates", value=new_missing_date_range, data_columns=True)


def _record_gaps(output_filename, pv_system_id, date_to, timeseries, datetime_of_api_request):
    dates_of_data = (
        timeseries["instantaneous_power_gen_W"].dropna().resample("D").mean().dropna().index.date
    )
    dates_requested = pd.date_range(date_to - timedelta(days=365), date_to, freq="D").date
    missing_dates = set(dates_requested) - set(dates_of_data)
    missing_date_ranges = _convert_consecutive_dates_to_date_ranges(list(missing_dates))
    _LOG.info(
        "system_id %d: %d missing date ranges found: \n%s",
        pv_system_id,
        len(missing_date_ranges),
        missing_date_ranges,
    )
    if len(missing_date_ranges) == 0:
        return
    # Convert to from date objects to pd.Timestamp objects, because HDF5
    # doesn't like to store date objects.
    missing_date_ranges = missing_date_ranges.astype("datetime64")
    missing_date_ranges["pv_system_id"] = pv_system_id
    missing_date_ranges["datetime_of_API_request"] = datetime_of_api_request
    missing_date_ranges.set_index("pv_system_id", inplace=True)
    with pd.HDFStore(output_filename, mode="a", complevel=9) as store:
        store.append(key="missing_dates", value=missing_date_ranges, data_columns=True)


def _convert_consecutive_dates_to_date_ranges(missing_dates):
    new_missing = []
    missing_dates = np.sort(np.unique(missing_dates))
    if len(missing_dates) == 0:
        return pd.DataFrame(new_missing)

    gaps = np.diff(missing_dates).astype("timedelta64[D]").astype(int) > 1
    gaps = np.where(gaps)[0]

    start_date = missing_dates[0]
    for gap_i in gaps:
        end_date = missing_dates[gap_i]
        new_missing.append(
            {
                "missing_start_date_PV_localtime": start_date,
                "missing_end_date_PV_localtime": end_date,
            }
        )
        start_date = missing_dates[gap_i + 1]

    end_date = missing_dates[-1]
    new_missing.append(
        {"missing_start_date_PV_localtime": start_date, "missing_end_date_PV_localtime": end_date}
    )

    return pd.DataFrame(new_missing)



"""Retrieve Natural Earth world boundaries."""

import logging
import os
from io import BytesIO
from typing import Tuple

import geopandas as gpd
import requests
from numpy.typing import NDArray


class NaturalEarth:
    """Retrieve Natural Earth world boundaries."""

    def __init__(self, cache_dir: str = None) -> None:
        """Initialise.

        Args:
            cache_dir:
                Optionally provide a location to cache boundary definition files locally and avoid
                unnecsessary downloads.

        Raises:
            ValueError: If the cache_dir does not exist.
        """
        self.cache_dir = cache_dir
        if self.cache_dir is not None:
            if not os.path.isdir(cache_dir):
                logging.error("The cache_dir does not exist.")
                raise ValueError("The cache_dir does not exist.")
        self.world_hires = None
        self.countries_hires = None
        self.world_lores = None
        self.countries_lores = None

    def get_hires_world_boundaries(self) -> Tuple[gpd.GeoDataFrame, NDArray]:
        """Load high res world boundaries.

        Download the high resolution country boundaries GIS file from the Natural Earth website and
        optionally cache locally.

        Returns:
            A tuple containing (`world`, `countries`). `world` is a Geopandas GeoDataFrame with
            geometries and metadata for all country borders. `countries` is a list of unique country
            names for which geometries exist. Boundaries will be in the EPSG:4087 projected CRS.

        Typical usage example:
            world, countries = get_world_boundaries()
        """
        if self.world_hires is not None and self.countries_hires is not None:
            return self.world_hires, self.countries_hires
        if self.cache_dir:
            cache_file = os.path.join(self.cache_dir, "ne_10m_admin_0_countries.zip")
        else:
            cache_file = None
        if cache_file is not None and os.path.isfile(cache_file):
            data = cache_file
        else:
            headers = {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
                "(KHTML, like Gecko) Chrome/95.0.4638.69 Safari/537.36"
            }
            url = (
                "https://www.naturalearthdata.com/http//www.naturalearthdata.com/"
                "download/10m/cultural/ne_10m_admin_0_countries.zip"
            )
            req = requests.get(url, headers=headers)
            data = BytesIO(req.content)
            if cache_file is not None:
                with open(cache_file, "wb") as fid:
                    fid.write(req.content)
        self.world_hires = gpd.read_file(data).to_crs("EPSG:4087")
        cols2keep = {"NAME": "name", "CONTINENT": "continent", "geometry": "geometry"}
        self.world_hires = self.world_hires[list(cols2keep.keys())].rename(columns=cols2keep)
        self.countries_hires = self.world_hires.name.unique()
        return self.world_hires, self.countries_hires

    def get_lores_world_boundaries(self) -> Tuple[gpd.GeoDataFrame, NDArray]:
        """Load low resolution world boundaries.

        Load the low res world boundaries GIS file (`naturalearth_lowres`) from Geopandas datasets.
        Useful for visualisations and/or to speed up computation.

        Returns:
            A tuple containing (`world`, `countries`). `world` is a Geopandas GeoDataFrame with
            geometries and metadata for all country borders. `countries` is a list of unique country
            names for which geometries exist. Boundaries will be in the EPSG:4087 projected CRS.

        Typical usage example:
            world, countries = get_world_boundaries()
        """
        if self.world_lores is not None and self.countries_lores is not None:
            return self.world_lores, self.countries_lores
        self.world_lores = gpd.read_file(gpd.datasets.get_path("naturalearth_lowres")).to_crs(
            "EPSG:4087"
        )
        self.world_lores.drop(columns=["pop_est", "iso_a3", "gdp_md_est"], inplace=True)
        self.countries = self.world_lores.name.unique()
        return self.world_lores, self.countries_lores

    def list_countries(self, res: str = "hires") -> Tuple[gpd.GeoDataFrame, NDArray]:
        """Print a list of country names to stdout.

        Print a list of country names whose geometries are available in the world boundaries GIS
        file.

        Args:
            res:
                Optionally switch between 'hires' and 'lores', although in theory both country lists
                should be identical, there may be some countries that are not included in the lores
                boundaries due to downsampling of borders. Names may also have changed between the
                two (e.g. Macedonia -> North Macedonia).

        Raises:
            ValueError: If `res` is not one of: 'lores', 'hires'.
        """
        if res == "hires":
            _, countries = self.get_hires_world_boundaries()
        elif res == "lores":
            _, countries = self.get_lores_world_boundaries()
        else:
            logging.error("The `res` arg should be one of: 'lores', 'hires'.")
            raise ValueError("The `res` arg should be one of: 'lores', 'hires'.")
        countries.sort()
        print(f"Available countries are:\n{', '.join(countries)}")

        return countries



from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
install_requires = (this_directory / "requirements.txt").read_text().splitlines()
long_description = (this_directory / "README.md").read_text()

setup(
    name="pvoutput-ocf",
    version="0.1.32",
    license="MIT",
    packages=find_packages(),
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    company="Open Climate Fix Ltd",
    author_email="info@openclimatefix.org",
)



from datetime import date, datetime
from io import StringIO

import numpy as np
import pandas as pd
import pytest

from pvoutput import pvoutput


def test_init():
    _ = pvoutput.PVOutput(api_key="fake", system_id="fake")


def test_rate_limit():
    pv = pvoutput.PVOutput(api_key="fake", system_id="fake")

    # set a fake reset time
    pv.rate_limit_reset_time = pd.Timestamp.utcnow() + pd.Timedelta(minutes=30)

    # get the number of seconds we need to wait
    seconds_to_wait = pv.wait_for_rate_limit_reset(do_sleeping=False)

    # 30 mins, + 3 mins for safety
    assert np.round(seconds_to_wait) == 30 * 60 + (60 * 3)


@pytest.mark.skip("Currently not working in CI")
def test_get_status():
    pv = pvoutput.PVOutput()
    pv.get_status(
        pv_system_id=10033,
        date=datetime(2022, 3, 1, 12),
        use_data_service=True,
        timezone="Europe/London",
    )


@pytest.mark.skip("Currently not working in CI")
def test_multiple_get_status():
    pv = pvoutput.PVOutput()
    status_df = pv.get_system_status(
        pv_system_ids=[10033, 10020],
        date=datetime(2022, 3, 15),
        use_data_service=True,
        timezone="Europe/London",
    )

    assert len(status_df) > 0


def test_convert_consecutive_dates_to_date_ranges():
    dr1 = pd.date_range("2018-01-01", "2018-02-01", freq="D").tolist()
    dr2 = pd.date_range("2018-02-05", "2018-02-10", freq="D").tolist()
    missing_dates = dr1 + dr2
    date_ranges = pvoutput._convert_consecutive_dates_to_date_ranges(missing_dates)
    columns = ["missing_start_date_PV_localtime", "missing_end_date_PV_localtime"]
    pd.testing.assert_frame_equal(
        date_ranges[columns],
        pd.DataFrame(
            [
                [dr1[0], dr1[-1]],
                [dr2[0], dr2[-1]],
            ],
            columns=columns,
        ),
    )


def test_date_to_pvoutput_str():
    VALID_DATE_STR = "20190101"
    assert pvoutput.date_to_pvoutput_str(VALID_DATE_STR) == VALID_DATE_STR
    ts = pd.Timestamp(VALID_DATE_STR)
    assert pvoutput.date_to_pvoutput_str(ts) == VALID_DATE_STR


def test_check_date():
    assert pvoutput._check_date("20190101") is None
    with pytest.raises(ValueError):
        pvoutput._check_date("2010")
    with pytest.raises(ValueError):
        pvoutput._check_date("2010-01-02")


def test_check_pv_system_status():
    def _make_timeseries(start, end):
        index = pd.date_range(start, end, freq="5T")
        n = len(index)
        timeseries = pd.DataFrame(np.zeros(n), index=index)
        return timeseries

    DATE = date(2019, 1, 1)
    good_timeseries = _make_timeseries("2019-01-01 00:00", "2019-01-02 00:00")
    pvoutput.check_pv_system_status(good_timeseries, DATE)

    bad_timeseries = _make_timeseries("2019-01-01 00:00", "2019-01-03 00:00")
    with pytest.raises(ValueError):
        pvoutput.check_pv_system_status(bad_timeseries, DATE)

    bad_timeseries2 = _make_timeseries("2019-01-02 00:00", "2019-01-03 00:00")
    with pytest.raises(ValueError):
        pvoutput.check_pv_system_status(bad_timeseries2, DATE)



"""Code for scraping for pv systems"""

import re
from copy import copy
from typing import Iterable, Optional, Union

import pandas as pd
import requests
from bs4 import BeautifulSoup

from pvoutput.consts import (
    MAP_URL,
    PV_OUTPUT_COUNTRY_CODES,
    PV_OUTPUT_MAP_COLUMN_NAMES,
    REGIONS_URL,
)

_MAX_NUM_PAGES = 1024


def get_pv_systems_for_country(
    country: Union[str, int],
    ascending: Optional[bool] = None,
    sort_by: Optional[str] = None,
    max_pages: int = _MAX_NUM_PAGES,
    region: Optional[str] = None,
) -> pd.DataFrame:
    """
    Get all pv systems for on country

    Args:
        country: either a string such as 'United Kingdom'
            (see consts.PV_OUTPUT_COUNTRY_CODES for all recognised strings),
            or a PVOutput.org country code, in the range [1, 257].
        ascending: if True, ask PVOutput.org to sort results by ascending.
            If False, sort by descending.  If None, use PVOutput.org's default
            sort order.
        sort_by: The column to ask PVOutput.org to sort by.  One of:
            timeseries_duration,
            average_generation_per_day,
            efficiency,
            power_generation,
            capacity,
            address,
            name
        max_pages: The maximum number of search pages to scrape.
        region: Optional input, #TODO

    Returns: pd.DataFrame with index system_id (int) and these columns:
        name, system_DC_capacity_W, panel, inverter, address, orientation,
        array_tilt_degrees, shade, timeseries_duration,
        total_energy_gen_Wh, average_daily_energy_gen_Wh
        average_efficiency_kWh_per_kW
    """
    country_code = _convert_to_country_code(country)
    regions = [region] if region else get_regions_for_country(country_code)
    all_metadata = []
    for region in regions:
        for page_number in range(max_pages):
            print(
                "\rReading page {:2d} for region: {}".format(page_number, region),
                end="",
                flush=True,
            )
            url = _create_map_url(
                country_code=country_code,
                page_number=page_number,
                ascending=ascending,
                sort_by=sort_by,
                region=region,
            )
            soup = get_soup(url)
            if _page_is_blank(soup):
                break
            metadata = _process_metadata(soup)
            metadata["region"] = region
            all_metadata.append(metadata)

            if not _page_has_next_link(soup):
                break

    return pd.concat(all_metadata)


# ########### LOAD HTML ###################


def _create_map_url(
    country_code: Optional[int] = None,
    page_number: Optional[int] = None,
    ascending: Optional[bool] = None,
    sort_by: Optional[str] = None,
    region: Optional[str] = None,
) -> str:
    """
    Create a map URL

    Args:
        country_code: Country code
        page_number: Get this page number of the search results.  Zero-indexed.
            The first page is page 0, the second page is page 1, etc.
        ascending: option for ascending or descending
        sort_by: sort results by (optional)
        region: region of country (optional)

    """
    _check_country_code(country_code)

    if ascending is None:
        sort_order = None
    else:
        sort_order = "asc" if ascending else "desc"

    if sort_by is None:
        sort_by_pv_output_col_name = None
    else:
        try:
            sort_by_pv_output_col_name = PV_OUTPUT_MAP_COLUMN_NAMES[sort_by]
        except KeyError:
            raise ValueError("sort_by must be one of {}".format(PV_OUTPUT_MAP_COLUMN_NAMES.keys()))

    url_params = {
        "country": country_code,
        "p": page_number,
        "d": sort_order,
        "o": sort_by_pv_output_col_name,
        "region": region,
    }

    url_params_list = [
        "{}={}".format(key, value) for key, value in url_params.items() if value is not None
    ]
    query_string = "&".join(url_params_list)
    url = copy(MAP_URL)
    if query_string:
        url += "?" + query_string
    return url


def _raise_country_error(country, msg=""):
    country_codes = PV_OUTPUT_COUNTRY_CODES.values()
    raise ValueError(
        "Wrong value country='{}'. {}country must be an integer country"
        " code in the range [{}, {}], or one of {}.".format(
            country,
            msg,
            min(country_codes),
            max(country_codes),
            ", ".join(PV_OUTPUT_COUNTRY_CODES.keys()),
        )
    )


def _check_country_code(country_code: Union[None, int]):
    if country_code is None:
        return
    country_codes = PV_OUTPUT_COUNTRY_CODES.values()
    if not min(country_codes) <= country_code <= max(country_codes):
        _raise_country_error(country_code, "country outside of valid range!  ")


def _convert_to_country_code(country: Union[str, int]) -> int:
    if isinstance(country, str):
        try:
            return PV_OUTPUT_COUNTRY_CODES[country]
        except KeyError:
            _raise_country_error(country)

    elif isinstance(country, int):
        _check_country_code(country)
        return country


def _page_has_next_link(soup: BeautifulSoup):
    return bool(soup.find_all("a", text="Next"))


# ############ PROCESS HTML #########################


def _process_metadata(soup: BeautifulSoup, return_constituents=False) -> pd.DataFrame:
    pv_system_size_metadata = _process_system_size_col(soup)
    index = pv_system_size_metadata.index
    pv_systems_metadata = [
        pv_system_size_metadata,
        _process_output_col(soup, index),
        _process_generation_and_average_cols(soup, index),
        _process_efficiency_col(soup, index),
    ]

    df = pd.concat(pv_systems_metadata, axis="columns")
    df = _convert_metadata_cols_to_numeric(df)
    df["system_DC_capacity_W"] = df["capacity_kW"] * 1e3
    del df["capacity_kW"]
    if return_constituents:
        pv_systems_metadata.append(df)
        return tuple(pv_systems_metadata)
    return df


def _process_system_size_col(soup: BeautifulSoup) -> pd.DataFrame:
    pv_system_size_col = soup.find_all("a", href=re.compile(r"display\.jsp\?sid="))
    metadata = []
    for row in pv_system_size_col:
        metadata_for_row = {}

        # Get system ID
        href = row.attrs["href"]
        p = re.compile(r"^display\.jsp\?sid=(\d+)$")
        href_match = p.match(href)
        metadata_for_row["system_id"] = href_match.group(1)

        # Process title (lots of metadata in here!)
        title, title_meta = row.attrs["title"].split("|")

        # Name and capacity
        p = re.compile(r"(.*) (\d+\.\d+kW)")
        title_match = p.match(title)
        metadata_for_row["name"] = title_match.group(1)
        metadata_for_row["capacity"] = title_match.group(2)

        # Other key-value pairs:
        key_value = title_meta.split("<br/>")
        key_value_dict = {}
        for line in key_value:
            key_value_split = line.split(":")
            key = key_value_split[0].strip()
            # Some values have a colon(!)
            value = ":".join(key_value_split[1:]).strip()
            key_value_dict[key] = value
        metadata_for_row.update(key_value_dict)

        # Some cleaning
        # Remove <img ...> from Location
        location = metadata_for_row["Location"]
        p = re.compile(r"(<img .*\>)?(.*)")
        img_groups = p.search(location).groups()
        if img_groups[0] is not None:
            metadata_for_row["Location"] = img_groups[1].strip()

        metadata.append(metadata_for_row)

    df = pd.DataFrame(metadata)
    df["system_id"] = pd.to_numeric(df["system_id"])
    df = df.set_index("system_id")
    df.columns = [col_name.lower() for col_name in df.columns]
    df.rename(
        {
            "location": "address",
            "panels": "panel",
            "array tilt": "array_tilt_degrees",
            "capacity": "capacity_kW",
        },
        axis="columns",
        inplace=True,
    )
    return df


def _remove_str_and_convert_to_numeric(series: pd.Series, string_to_remove: str) -> pd.Series:
    series = series.str.replace(string_to_remove, "")
    return pd.to_numeric(series)


def _convert_metadata_cols_to_numeric(df: pd.DataFrame) -> pd.DataFrame:
    for col_name, string_to_remove in [
        # ('array_tilt_degrees', '°'),
        ("capacity_kW", "kW"),
        ("average_efficiency_kWh_per_kW", "kWh/kW"),
    ]:
        df[col_name] = _remove_str_and_convert_to_numeric(df[col_name], string_to_remove)

    return df


def _process_output_col(soup: BeautifulSoup, index: Optional[Iterable] = None) -> pd.Series:

    # get all data
    outputs_col = soup.find_all(text=re.compile(r"\d Days"))

    # format data as strings
    outputs_col = [str(col) for col in outputs_col]

    # make into pandas Series
    duration = pd.Series(outputs_col, name="timeseries_duration", index=index)

    # change to timedeltas
    return pd.to_timedelta(duration.astype("unicode"))


def _convert_energy_to_numeric_watt_hours(series: pd.Series) -> pd.Series:
    data = []
    for unit, multiplier in [("kWh", 1e3), ("MWh", 1e6)]:
        selection = series[series.str.contains(unit)]
        selection = selection.str.replace(unit, "")
        selection = selection.str.replace(",", "")
        selection = pd.to_numeric(selection)
        selection *= multiplier
        data.append(selection)
    return pd.concat(data)


def _process_generation_and_average_cols(
    soup: BeautifulSoup, index: Optional[Iterable] = None
) -> pd.DataFrame:
    # _soup = deepcopy(soup)
    [s.decompose() for s in soup.select("a")]
    generation_and_average_cols = soup.find_all(text=re.compile(r"\d[Mk]Wh$"))
    generation_col = generation_and_average_cols[0::2]
    average_col = generation_and_average_cols[1::2]
    df = pd.DataFrame(
        {"total_energy_gen_Wh": generation_col, "average_daily_energy_gen_Wh": average_col},
        index=index,
    )

    for col_name in df.columns:
        df[col_name] = _convert_energy_to_numeric_watt_hours(df[col_name])

    return df


def _process_efficiency_col(soup: BeautifulSoup, index: Optional[Iterable] = None) -> pd.Series:
    efficiency_col = soup.find_all(text=re.compile(r"\dkWh/kW"))
    return pd.Series(efficiency_col, name="average_efficiency_kWh_per_kW", index=index)


def _page_is_blank(soup: BeautifulSoup) -> bool:
    # Pages can still be blank even if the previous page has a Next Button
    pv_system_size_col = soup.find_all("a", href=re.compile(r"display\.jsp\?sid="))
    return not bool(pv_system_size_col)


def get_soup(url, raw=False, parser="html.parser"):
    """
    Get soupt from url

    Args:
        url: URL
        raw: option for raw, defaulted to False
        parser: parser for BeautifulSoup

    """
    response = requests.get(url)
    soup = BeautifulSoup(response.text, parser)
    if raw:
        return soup
    return clean_soup(soup)


def clean_soup(soup):
    """Function to clean scraped soup object.

    Note that the downloaded soup could change over time.
    Args:
        soup: bs4.BeautifulSoup

    Returns:
        bs4.BeautifulSoup

    """
    for script in soup.find_all("script", src=False):
        script.decompose()
    return soup


def get_regions_for_country(country_code: int):
    """
    Get regions for on countruy

    Args:
        country_code: the country code

    Returns: list of regions
    """
    region_list = []
    url = f"{REGIONS_URL}?country={country_code}"
    soup = get_soup(url, parser="lxml")
    region_tags = soup.find_all("a", href=re.compile(r"map\.jsp\?country="))
    for row in region_tags:
        href = row.attrs["href"]
        p = re.compile(r"^map\.jsp\?country=" + str(country_code) + r"&region=(\w+.*)$")
        href_match = p.match(href)
        region = href_match.group(1)
        region_list.append(region)
    return region_list
